@@ -12,7 +12,6 @@
 #include "ld_hash.h"
 #include "ld_string.h"
 
-extern int _dl_error_number;
 extern struct r_debug *_dl_debug_addr;
 
 extern void *(*_dl_malloc_function) (size_t size);
@@ -26,6 +25,7 @@ void *dlsym(void *, const char *) __attribute__ ((__weak__, __alias__ ("_dlsym")
 int dlclose(void *) __attribute__ ((__weak__, __alias__ ("_dlclose")));
 int dladdr(void *, Dl_info *) __attribute__ ((__weak__, __alias__ ("_dladdr")));
 
+#ifdef __PIC__
 /* This is a real hack.  We need access to the dynamic linker, but we
 also need to make it possible to link against this library without any
 unresolved externals.  We provide these weak symbols to make the link
@@ -47,6 +47,10 @@ extern int _dl_parse_relocation_information(struct elf_resolve *, unsigned long,
 	__attribute__ ((__weak__, __alias__ ("foobar")));
 extern void _dl_parse_lazy_relocation_information(struct elf_resolve *, unsigned long, unsigned long, int)
 	__attribute__ ((__weak__, __alias__ ("foobar")));
+#ifdef __mips__
+extern void _dl_perform_mips_global_got_relocations(struct elf_resolve *tpnt)
+	__attribute__ ((__weak__, __alias__ ("foobar")));
+#endif
 #ifdef USE_CACHE
 int _dl_map_cache(void) __attribute__ ((__weak__, __alias__ ("foobar")));
 int _dl_unmap_cache(void) __attribute__ ((__weak__, __alias__ ("foobar")));
@@ -56,8 +60,29 @@ extern struct dyn_elf *_dl_symbol_tables __attribute__ ((__weak__, __alias__ ("f
 extern struct dyn_elf *_dl_handles __attribute__ ((__weak__, __alias__ ("foobar1")));
 extern struct elf_resolve *_dl_loaded_modules __attribute__ ((__weak__, __alias__ ("foobar1")));
 extern struct r_debug *_dl_debug_addr __attribute__ ((__weak__, __alias__ ("foobar1")));
-extern int _dl_error_number __attribute__ ((__weak__, __alias__ ("foobar1")));
+extern unsigned long _dl_error_number __attribute__ ((__weak__, __alias__ ("foobar1")));
 extern void *(*_dl_malloc_function)(size_t) __attribute__ ((__weak__, __alias__ ("foobar1")));
+#else
+#ifdef __SUPPORT_LD_DEBUG__
+static char *_dl_debug  = 0;
+static char *_dl_debug_symbols = 0;
+static char *_dl_debug_move    = 0;
+static char *_dl_debug_reloc   = 0;
+static char *_dl_debug_detail  = 0;
+static char *_dl_debug_nofixups  = 0;
+static char *_dl_debug_bindings  = 0;
+static int   _dl_debug_file = 2;
+#elif defined __SUPPORT_LD_DEBUG_EARLY__
+#define _dl_debug_file 2
+#endif
+char *_dl_library_path = 0;
+char *_dl_ldsopath = 0;
+struct r_debug *_dl_debug_addr = NULL;
+static char *_dl_malloc_addr, *_dl_mmap_zero;
+#include "../ldso/ldso.h"               /* Pull in the name of ld.so */
+#include "../ldso/hash.c"
+#include "../ldso/readelflib1.c"
+#endif
 
 static const char *dl_error_names[] = {
 	"",
@@ -105,7 +130,15 @@ void *_dlopen(const char *libname, int flag)
 	static int dl_init = 0;
 	char *from;
 	void (*dl_brk) (void);
+#ifdef __PIC__
 	int (*dl_elf_init) (void);
+#endif
+
+	/* A bit of sanity checking... */
+	if (!(flag & (RTLD_LAZY|RTLD_NOW))) {
+		_dl_error_number = LD_BAD_HANDLE;
+		return NULL;
+	}
 
 	from = __builtin_return_address(0);
 
@@ -145,9 +178,10 @@ void *_dlopen(const char *libname, int flag)
 #endif
 		return NULL;
 	}
+	//tpnt->libtype = loaded_file;
 
 	dyn_chain = rpnt = (struct dyn_elf *) malloc(sizeof(struct dyn_elf));
-	_dl_memset(rpnt, 0, sizeof(*rpnt));
+	_dl_memset(rpnt, 0, sizeof(struct dyn_elf));
 	rpnt->dyn = tpnt;
 	rpnt->flags = flag;
 	if (!tpnt->symbol_scope)
@@ -179,7 +213,7 @@ void *_dlopen(const char *libname, int flag)
 			goto oops;
 
 		      rpnt->next = (struct dyn_elf *) malloc(sizeof(struct dyn_elf));
-		      _dl_memset (rpnt->next, 0, sizeof (*(rpnt->next)));
+		      _dl_memset (rpnt->next, 0, sizeof (struct dyn_elf));
 		      rpnt = rpnt->next;
 		      if (!tpnt1->symbol_scope) tpnt1->symbol_scope = dyn_chain;
 		      rpnt->dyn = tpnt1;
@@ -196,25 +230,41 @@ void *_dlopen(const char *libname, int flag)
 
 	rpnt->next = _dl_symbol_tables;
 
+	/*
+	 * MIPS is special *sigh*
+	 */
+#ifdef __mips__
+	_dl_perform_mips_global_got_relocations(tpnt);
+#endif
+
 	if (do_fixup(tpnt, flag)) {
 		_dl_error_number = LD_NO_SYMBOL;
 		goto oops;
 	}
 
-	dl_brk = (void (*)(void)) _dl_debug_addr->r_brk;
-	if (dl_brk != NULL) {
+	if (_dl_debug_addr) {
+	    dl_brk = (void (*)(void)) _dl_debug_addr->r_brk;
+	    if (dl_brk != NULL) {
 		_dl_debug_addr->r_state = RT_ADD;
 		(*dl_brk) ();
 
 		_dl_debug_addr->r_state = RT_CONSISTENT;
 		(*dl_brk) ();
+	    }
 	}
 
-	for (rpnt = dyn_chain; rpnt; rpnt = rpnt->next) {
-		tpnt = rpnt->dyn;
+#ifdef __PIC__
+	/* Find the last library */
+	for (tpnt = dyn_chain->dyn; tpnt->next!=NULL; tpnt = tpnt->next)
+		;
+	/* Run the ctors and set up the dtors */
+	for (; tpnt != dyn_chain->dyn->prev; tpnt=tpnt->prev)
+	{
 		/* Apparently crt1 for the application is responsible for handling this.
 		 * We only need to run the init/fini for shared libraries
 		 */
+		if (tpnt->libtype == program_interpreter)
+			continue;
 		if (tpnt->libtype == elf_executable)
 			continue;
 		if (tpnt->init_flag & INIT_FUNCS_CALLED)
@@ -230,6 +280,7 @@ void *_dlopen(const char *libname, int flag)
 		}
 
 	}
+#endif
 
 #ifdef USE_CACHE
 	_dl_unmap_cache();
@@ -281,14 +332,15 @@ static int do_fixup(struct elf_resolve *tpnt, int flag)
 			return goof;
 		tpnt->init_flag |= JMP_RELOCS_DONE;
 
-		if (flag == RTLD_LAZY)
+		if (flag == RTLD_LAZY) {
 			_dl_parse_lazy_relocation_information(tpnt, 
 				tpnt->dynamic_info[DT_JMPREL], 
 				tpnt->dynamic_info[DT_PLTRELSZ], 0);
-		else
+		} else {
 			goof += _dl_parse_relocation_information(tpnt, 
 				tpnt->dynamic_info[DT_JMPREL], 
 				tpnt->dynamic_info[DT_PLTRELSZ], 0);
+		}
 	};
 	return goof;
 }
@@ -462,13 +514,15 @@ static int do_dlclose(void *vhandle, int need_fini)
 	}
 
 
-	dl_brk = (void (*)(void)) _dl_debug_addr->r_brk;
-	if (dl_brk != NULL) {
+	if (_dl_debug_addr) {
+	    dl_brk = (void (*)(void)) _dl_debug_addr->r_brk;
+	    if (dl_brk != NULL) {
 		_dl_debug_addr->r_state = RT_DELETE;
 		(*dl_brk) ();
 
 		_dl_debug_addr->r_state = RT_CONSISTENT;
 		(*dl_brk) ();
+	    }
 	}
 
 	return 0;
