@@ -36,6 +36,35 @@
  *     and error state.
  *  The above changes should take care of a problem initially reported
  *  by "Steven J. Hill" <sjhill@realitydiluted.com>.
+ *
+ *  8-25-2002
+ *  Changed fclose behavior when custom streams were enabled.  Previously,
+ *     the cookie pointer was set to NULL as a debugging aid.  However,
+ *     some of the perl 5.8 test rely on being able to close stderr and
+ *     still try writing to it.  So now, the cookie pointer and handler
+ *     function pointers are set to that it is a "normal" file with a
+ *     file descriptor of -1.  Note: The cookie pointer is reset to NULL
+ *     if the FILE struct is free'd by fclose.
+ *
+ *  Nov 21, 2002
+ *  Added internal function _wstdio_fwrite.
+ *  Jan 3, 2003
+ *  Fixed a bug in _wstdio_fwrite.
+ *
+ *  Jan 22, 2003
+ *  Fixed a bug related file position in append mode.  _stdio_fwrite now
+ *     seeks to the end of the stream when append mode is set and we are
+ *     transitioning to write mode, so that subsequent ftell() return
+ *     values are correct.
+ *  Also fix _stdio_fopen to support fdopen() with append specified when
+ *     the underlying file didn't have O_APPEND set.  It now sets the
+ *     O_APPEND flag as recommended by SUSv3 and is done by glibc.
+ *
+ *  May 15, 2003
+ *  Modify __stdio_fread to deal with fake streams used by *sscanf.
+ *  Set EOF to end of buffer when fmemopen used on a readonly stream.
+ *    Note: I really need to run some tests on this to see what the
+ *    glibc code does in each case.
  */
 
 /* Before we include anything, convert L_ctermid to L_ctermid_function
@@ -307,11 +336,11 @@ int putw(int w, FILE *stream)
 UNLOCKED(int,fileno,(register FILE *stream),(stream))
 {
 #ifdef __STDIO_GLIBC_CUSTOM_STREAMS
-	return ( ((stream->cookie == &(stream->filedes)) && (stream->filedes >= 0))
+	return ( (stream && (stream->cookie == &(stream->filedes)) && (stream->filedes >= 0))
 			 ? stream->filedes
 			 : (__set_errno(EBADF), -1) );
 #else  /* __STDIO_GLIBC_CUSTOM_STREAMS */
-	return (stream->filedes >= 0) ? stream->filedes : (__set_errno(EBADF), -1);
+	return ((stream && stream->filedes >= 0)) ? stream->filedes : (__set_errno(EBADF), -1);
 #endif /* __STDIO_GLIBC_CUSTOM_STREAMS */
 }
 
@@ -616,6 +645,9 @@ FILE *fmemopen(void *s, size_t len, const char *modes)
 
 		if (fp != NULL) {
 			cookie->fp = fp;
+			if (fp->modeflags & __FLAG_READONLY) {
+				cookie->eof = len;
+			}
 			if ((fp->modeflags & __FLAG_APPEND) && (len > 0)) {
 				for (i = 0 ; i < len ; i++) {
 					if (cookie->buf[i] == 0) {
@@ -1023,31 +1055,43 @@ void _flushlbf(void)
 /**********************************************************************/
 #ifdef L___fsetlocking
 
-/* No (serious) reentrancy issues -- return value could be incorrect. */
-/* TODO -- fix race */
+/* NOT threadsafe!!!  (I don't think glibc's is either)
+ *
+ * This interacts badly with internal locking/unlocking.  If you use this routine,
+ * make sure the file isn't being accessed by any other threads.  Typical use would
+ * be to change to user locking immediately after opening the stream.
+ */
+
+#ifdef __UCLIBC_MJN3_ONLY__
+link_warning(__fsetlocking, "Oddly enough, __fsetlocking() is NOT threadsafe.")
+#endif
 
 int __fsetlocking(FILE *stream, int locking_mode)
 {
+#ifdef __STDIO_THREADSAFE
 	int old_mode;
+#endif
 
 	assert((FSETLOCKING_QUERY == 0) && (FSETLOCKING_INTERNAL == 1)
 		   && (FSETLOCKING_BYCALLER == 2));
 
 	assert(((unsigned int) locking_mode) <= 2);
 
-	/* Note: don't even bother locking here... */
+#ifdef __STDIO_THREADSAFE
+	old_mode = stream->user_locking;
 
-	old_mode  = stream->user_locking;
+	assert(((unsigned int) old_mode) <= 1);	/* Must be 0 (internal) or 1 (user). */
 
 	if (locking_mode != FSETLOCKING_QUERY) {
 		/* In case we're not debugging, treat any unknown as a request to
 		 * set internal locking, in order to match glibc behavior. */
-		stream->user_locking = ((locking_mode == FSETLOCKING_BYCALLER)
-								? FSETLOCKING_BYCALLER
-								: FSETLOCKING_INTERNAL);
+		stream->user_locking = (locking_mode == FSETLOCKING_BYCALLER);
 	}
 
-	return old_mode;
+	return 2 - old_mode;
+#else
+	return FSETLOCKING_BYCALLER; /* Well, without thread support... */
+#endif
 }
 
 #endif
@@ -1371,7 +1415,7 @@ size_t _stdio_fread(unsigned char *buffer, size_t bytes, register FILE *stream)
 			*p++ = *stream->bufpos++;
 		}
 
-		if (bytes > 0) {
+		if ((bytes > 0) && (stream->filedes != -2)) {
 			ssize_t len;
 
 			/* The buffer is exhausted, but we still need chars.  */
@@ -1481,8 +1525,8 @@ size_t _stdio_fread(unsigned char *buffer, size_t bytes, register FILE *stream)
  * error flag before/after the write process, but it doesn't seem worth
  * the trouble. */
 
-/* Like standard write, but always does a full write unless error plus
- *deals correctly with bufsize > SSIZE_MAX... not much on an issue on linux
+/* Like standard write, but always does a full write unless error, plus
+ * deals correctly with bufsize > SSIZE_MAX... not much on an issue on linux
  * but definitly could be on Elks.  Also on Elks, always loops for EINTR..
  * Returns number of bytes written, so a short write indicates an error */
 static size_t _stdio_WRITE(register FILE *stream,
@@ -1566,8 +1610,18 @@ size_t _stdio_fwrite(const unsigned char *buffer, size_t bytes,
 		stream->bufgetc =
 #endif /* __STDIO_GETC_MACRO */
 		stream->bufpos = stream->bufread = stream->bufstart;
-	}
+	} else
 #endif
+	if ((stream->modeflags & (__FLAG_WRITING|__FLAG_APPEND)) == __FLAG_APPEND) {
+		/* Append mode, but not currently writing.  Need to seek to end for proper
+		 * ftell() return values.  Don't worry if the stream isn't seekable. */
+		__offmax_t pos[1];
+		*pos = 0;
+		if (_stdio_lseek(stream, pos, SEEK_END) && (errno != EPIPE)) { /* Too big? */
+			stream->modeflags |= __FLAG_ERROR;
+			return 0;
+		}
+	}
 
 #ifdef __STDIO_PUTC_MACRO
 	/* We need to disable putc macro in case of error */
@@ -1690,6 +1744,7 @@ size_t _stdio_fwrite(const unsigned char *buffer, size_t bytes,
 
 	/* We always clear the reading flag in case at EOF. */
 	stream->modeflags &= ~(__FLAG_READING);
+
 	/* Unlike the buffered case, we set the writing flag now since we don't
 	 * need to do anything here for fflush(). */
 	stream->modeflags |= __FLAG_WRITING;
@@ -1851,7 +1906,7 @@ void _stdio_term(void)
 	 * Note: Set locking mode to "by caller" to save some overhead later. */
 	__stdio_init_mutex(&_stdio_openlist_lock);
 	for (ptr = _stdio_openlist ; ptr ; ptr = ptr->nextopen ) {
-		ptr->user_locking = FSETLOCKING_BYCALLER;
+		ptr->user_locking = 1;
 		__stdio_init_mutex(&ptr->lock);
 	}
 #endif /* __STDIO_THREADSAFE */
@@ -1898,9 +1953,11 @@ void _stdio_term(void)
 void _stdio_init(void)
 {
 #ifdef __STDIO_BUFFERS
+	int old_errno = errno;
 	/* stdin and stdout uses line buffering when connected to a tty. */
 	_stdio_streams[0].modeflags ^= (1-isatty(0)) * __FLAG_LBF;
 	_stdio_streams[1].modeflags ^= (1-isatty(1)) * __FLAG_LBF;
+	__set_errno(old_errno);
 #endif /* __STDIO_BUFFERS */
 #ifndef __UCLIBC__
 /* __stdio_term is automatically when exiting if stdio is used.
@@ -2009,7 +2066,11 @@ int fclose(register FILE *stream)
 		rv = EOF;
 	}
 #ifdef __STDIO_GLIBC_CUSTOM_STREAMS
-	stream->cookie = NULL;		/* To aid debugging... */
+	stream->cookie = &(stream->filedes);
+	stream->gcs.read = _cs_read;
+	stream->gcs.write = _cs_write;
+	stream->gcs.seek = 0;		/* The internal seek func handles normals. */
+	stream->gcs.close = _cs_close;
 #endif
 	stream->filedes = -1;		/* To aid debugging... */
 
@@ -2025,6 +2086,7 @@ int fclose(register FILE *stream)
 	/* At this point, any dangling refs to the stream are the result of
 	 * a programming bug... so free the unlocked stream. */
 	if (stream->modeflags & __FLAG_FREEFILE) {
+	    stream->cookie = NULL;	/* To aid debugging... */
 		free(stream);
 	}
 
@@ -2043,7 +2105,11 @@ int fclose(register FILE *stream)
 	}
 
 #ifdef __STDIO_GLIBC_CUSTOM_STREAMS
-	stream->cookie = NULL;		/* To aid debugging... */
+	stream->cookie = &(stream->filedes);
+	stream->gcs.read = _cs_read;
+	stream->gcs.write = _cs_write;
+	stream->gcs.seek = 0;		/* The internal seek func handles normals. */
+	stream->gcs.close = _cs_close;
 #endif
 	stream->filedes = -1;		/* To aid debugging... */
 
@@ -2052,6 +2118,7 @@ int fclose(register FILE *stream)
 	/* At this point, any dangling refs to the stream are the result of
 	 * a programming bug... so free the unlocked stream. */
 	if (stream->modeflags & __FLAG_FREEFILE) {
+	    stream->cookie = NULL;	/* To aid debugging... */
 		free(stream);
 	}
 
@@ -2145,14 +2212,18 @@ int fflush_unlocked(register FILE *stream)
 		if (_stdio_fwrite(NULL, 0, stream) > 0) { /* flush buffer contents. */
 			rv = -1;			/* Not all chars written. */
 		}
-	} else if (stream->modeflags & __FLAG_READONLY) {
-		/* According to info, glibc returns an error when the file is opened
-		 * in read-only mode.
-		 * ANSI/ISO says behavior in this case is undefined but also says you
-		 * shouldn't flush a stream you were reading from. */
-		stream->modeflags |= __FLAG_ERROR; /* TODO - check glibc behavior */
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning add option to test for undefined behavior of fflush
+#endif /* __UCLIBC_MJN3_ONLY__ */
+#if 0
+	} else if (stream->modeflags & (__FLAG_READING|__FLAG_READONLY)) {
+		/* ANSI/ISO says behavior in this case is undefined but also says you
+		 * shouldn't flush a stream you were reading from.  As usual, glibc
+		 * caters to broken programs and simply ignores this. */
+		stream->modeflags |= __FLAG_ERROR;
 		__set_errno(EBADF);
 		rv = -1;
+#endif
 	}
 
 #ifndef NDEBUG
@@ -2170,11 +2241,17 @@ int fflush_unlocked(register FILE *stream)
 	}
 #endif
 
-	/* TODO -- check glibc behavior regarding error indicator */
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning add option to test for undefined behavior of fflush
+#endif /* __UCLIBC_MJN3_ONLY__ */
+#if 0
 	return ((stream != NULL)
-			&& (stream->modeflags & __FLAG_READONLY)
+			&& (stream->modeflags & (__FLAG_READING|__FLAG_READONLY))
 			? ((stream->modeflags |= __FLAG_ERROR), __set_errno(EBADF), EOF)
 			: 0 );
+#else
+	return 0;
+#endif
 
 #endif /* __STDIO_BUFFERS */
 }
@@ -2286,17 +2363,26 @@ FILE *_stdio_fopen(const char * __restrict filename,
 	}
 
 	if (filedes >= 0) {			/* Handle fdopen trickery. */
-		/*
-		 * NOTE: it is insufficient to just check R/W/RW agreement.
-		 * We must also check for append mode agreement, as well as
-		 * largefile agreement if applicable.
-		 */
-		int i = (open_mode & (O_ACCMODE|O_APPEND|O_LARGEFILE)) + 1;
+		/* NOTE: it is insufficient to just check R/W/RW agreement.
+		 * We must also check largefile compatibility if applicable.
+		 * Also, if append mode is desired for fdopen but O_APPEND isn't
+		 * currently set, then set it as recommended by SUSv3.  However,
+		 * if append mode is not specified for fdopen but O_APPEND is set,
+		 * leave it set (glibc compat). */
+		int i = (open_mode & (O_ACCMODE|O_LARGEFILE)) + 1;
 
-		if ((i & (((int) filename) + 1)) != i) {
+		if (((i & (((int) filename) + 1)) != i)	/* Check basic agreement. */
+			|| (((open_mode & O_APPEND)
+				 && !(((int) filename) & O_APPEND)
+				 && fcntl(filedes, F_SETFL, O_APPEND)))	/* Need O_APPEND. */
+			) {
 			__set_errno(EINVAL);
 			filedes = -1;
 		}
+#ifdef __STDIO_LARGE_FILES
+		/* For later... to reflect largefile setting in stream flags. */
+		open_mode |= (((int) filename) & O_LARGEFILE);
+#endif /* __STDIO_LARGE_FILES */
 		stream->filedes = filedes;
 	} else {
 #ifdef __STDIO_LARGE_FILES
@@ -2320,10 +2406,16 @@ FILE *_stdio_fopen(const char * __restrict filename,
 		return NULL;
 	}
 
-	stream->modeflags |=
 #ifdef __STDIO_BUFFERS
-		(isatty(stream->filedes) * __FLAG_LBF) |
-#endif /* __STDIO_BUFFERS */
+	{
+	    /* Do not let isatty mess up errno */
+	    int old_errno = errno;
+	    stream->modeflags |= (isatty(stream->filedes) * __FLAG_LBF);
+	    __set_errno(old_errno);
+	}
+#endif
+
+	stream->modeflags |=
 #if (O_APPEND == __FLAG_APPEND) \
 && ((O_LARGEFILE == __FLAG_LARGEFILE) || (O_LARGEFILE == 0))
 		(open_mode & (O_APPEND|O_LARGEFILE)) | /* i386 linux and elks */
@@ -2358,6 +2450,7 @@ FILE *_stdio_fopen(const char * __restrict filename,
 #endif /* __STDIO_MBSTATE */
 
 #ifdef __STDIO_THREADSAFE
+	stream->user_locking = 0;
 	__stdio_init_mutex(&stream->lock);
 #endif /* __STDIO_THREADSAFE */
 
@@ -3278,6 +3371,15 @@ void _stdio_fdout(int fd, ...)
 #define INTERNAL_DIV_MOD
 #endif
 
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning REMINDER: move _uintmaxtostr to locale.c???
+#endif
+#include <locale.h>
+
+#ifndef __LOCALE_C_ONLY
+#define CUR_LOCALE			(__global_locale)
+#endif /* __LOCALE_C_ONLY */
+
 char *_uintmaxtostr(register char * __restrict bufend, uintmax_t uval,
 					int base, __UIM_CASE alphacase)
 {
@@ -3286,6 +3388,11 @@ char *_uintmaxtostr(register char * __restrict bufend, uintmax_t uval,
 #ifdef INTERNAL_DIV_MOD
 	unsigned int H, L, high, low, rh;
 #endif
+#ifndef __LOCALE_C_ONLY
+	int grouping, outdigit;
+	size_t gslen;		   /* This does not need to be initialized. */
+	const char *g;		   /* This does not need to be initialized. */
+#endif /* __LOCALE_C_ONLY */
 
 	negative = 0;
 	if (base < 0) {				/* signed value */
@@ -3299,14 +3406,66 @@ char *_uintmaxtostr(register char * __restrict bufend, uintmax_t uval,
 	/* this is an internal routine -- we shouldn't need to check this */
 	assert(!((base < 2) || (base > 36)));
 
+#ifndef __LOCALE_C_ONLY
+	grouping = -1;
+	outdigit = 0x80 & alphacase;
+	alphacase ^= outdigit;
+#ifdef __UCLIBC_MJN3_ONLY_
+#warning implement outdigit... need digit lengths!  (put it in locale struct)
+#endif
+	if (alphacase == __UIM_GROUP) {
+		assert(base == 10);
+		if (*(g = CUR_LOCALE.grouping)
+			&& ((gslen = strlen(CUR_LOCALE.thousands_sep)) > 0)
+			) {
+			grouping = *g;
+		}
+	}
+#endif /* __LOCALE_C_ONLY */
+
     *bufend = '\0';
 
 #ifndef INTERNAL_DIV_MOD
     do {
+#ifndef __LOCALE_C_ONLY
+		if (!grouping) {		/* Finished a group. */
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning REMINDER: decide about memcpy in _uintmaxtostr
+#endif
+#if 0
+			bufend -= gslen;
+			memcpy(bufend, CUR_LOCALE.thousands_sep, gslen);
+#else
+			grouping = gslen;
+			do {
+				*--bufend = CUR_LOCALE.thousands_sep[--grouping];
+			} while (grouping);
+#endif
+			if (g[1] != 0) { 	/* g[1] == 0 means repeat last grouping. */
+				/* Note: g[1] == -1 means no further grouping.  But since
+				 * we'll never wrap around, we can set grouping to -1 without
+				 * fear of */
+				++g;
+			}
+			grouping = *g;
+		}
+		--grouping;
+#endif /* __LOCALE_C_ONLY */
 		digit = uval % base;
 		uval /= base;
 
-		*--bufend = ( (digit < 10) ? digit + '0' : digit + alphacase );
+#ifndef __LOCALE_C_ONLY
+		if (outdigit) {
+			outdigit = CUR_LOCALE.outdigit_length[digit];
+			do {
+				*--bufend = (&CUR_LOCALE.outdigit0_mb)[digit][--outdigit];
+			} while (outdigit);
+			outdigit = 1;
+		} else
+#endif
+		{
+			*--bufend = ( (digit < 10) ? digit + '0' : digit + alphacase );
+		}
     } while (uval);
 
 #else  /* ************************************************** */
@@ -3321,13 +3480,49 @@ char *_uintmaxtostr(register char * __restrict bufend, uintmax_t uval,
 	high = (unsigned int) (uval >> (sizeof(unsigned int) * CHAR_BIT));
 
     do {
+#ifndef __LOCALE_C_ONLY
+		if (!grouping) {		/* Finished a group. */
+#ifdef __UCLIBC_MJN3_ONLY__
+#warning REMINDER: decide about memcpy in _uintmaxtostr
+#endif
+#if 0
+			bufend -= gslen;
+			memcpy(bufend, CUR_LOCALE.thousands_sep, gslen);
+#else
+			grouping = gslen;
+			do {
+				*--bufend = CUR_LOCALE.thousands_sep[--grouping];
+			} while (grouping);
+#endif
+			if (g[1] != 0) { 	/* g[1] == 0 means repeat last grouping. */
+				/* Note: g[1] == -1 means no further grouping.  But since
+				 * we'll never wrap around, we can set grouping to -1 without
+				 * fear of */
+				++g;
+			}
+			grouping = *g;
+		}
+		--grouping;
+#endif /* __LOCALE_C_ONLY */
+
 		rh = high % base;
 		high /= base;
 		digit = (low % base) + (L * rh);
 		low = (low / base) + (H * rh) + (digit / base);
 		digit %= base;
 		
-		*--bufend = ( (digit < 10) ? digit + '0' : digit + alphacase );
+#ifndef __LOCALE_C_ONLY
+		if (outdigit) {
+			outdigit = CUR_LOCALE.outdigit_length[digit];
+			do {
+				*--bufend = (&CUR_LOCALE.outdigit0_mb)[digit][--outdigit];
+			} while (outdigit);
+			outdigit = 1;
+		} else
+#endif
+		{
+			*--bufend = ( (digit < 10) ? digit + '0' : digit + alphacase );
+		}
     } while (low | high);
 
 #endif /******************************************************/
@@ -3342,3 +3537,58 @@ char *_uintmaxtostr(register char * __restrict bufend, uintmax_t uval,
 
 #endif
 /**********************************************************************/
+#ifdef L__wstdio_fwrite
+
+#include <wchar.h>
+
+size_t _wstdio_fwrite(const wchar_t *__restrict ws, size_t n,
+					  register FILE *__restrict stream)
+{
+	size_t r, count;
+	char buf[64];
+	const wchar_t *pw;
+
+#ifdef __STDIO_BUFFERS
+	if (stream->filedes == -3) { /* Special case to support {v}swprintf. */
+		count = ((wchar_t *)(stream->bufend)) - ((wchar_t *)(stream->bufpos));
+		if (count > n) {
+			count = n;
+		}
+		if (count) {
+			wmemcpy((wchar_t *)(stream->bufpos), ws, count);
+			stream->bufpos = (char *)(((wchar_t *)(stream->bufpos)) + count);
+		}
+		return n;
+	}
+#endif
+
+	if (stream->modeflags & __FLAG_NARROW) {
+		stream->modeflags |= __FLAG_ERROR;
+		__set_errno(EBADF);
+		return 0;
+	}
+	stream->modeflags |= __FLAG_WIDE;
+
+	pw = ws;
+	count = 0;
+	while (n > count) {
+		r = wcsnrtombs(buf, &pw, n-count, sizeof(buf), &stream->state);
+		if (r != ((size_t) -1)) { /* No encoding errors */
+			if (!r) {
+				++r;		  /* 0 is returned when nul is reached. */
+				pw = ws + count + r; /* pw was set to NULL, so correct. */
+			}
+			if (_stdio_fwrite(buf, r, stream) == r) {
+				count = pw - ws;
+				continue;
+			}
+		}
+		break;
+	}
+
+	/* Note: The count is incorrect if 0 < _stdio_fwrite return < r!!! */
+	return count;
+}
+
+#endif
+/**********************************************************************/
